// ZEVA C-series Motor Controller: Code for helper microcontroller
// Open Source version, released under MIT License (see Readme file)
// Last modified by Ian Hooper (ZEVA), October 2021

// Code for ATtiny13A
// Fuses: Standard except CKDIV8 off (giving 9.6Mhz internal clock)

// The helper micro shares ground reference with battery negative, so is galvanically
// isolated from the main microcontroller, which shares ground with the 12V system.
// Its only functions are desat (overcurrent) detection and measuring / transmitting
// battery voltage, passed to the main micro via high speed optocoupler.

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/wdt.h>

#define DESAT_SIGNAL	(1<<PB3)
#define PWM_PIN			(1<<PB4)

#define ADC_VREF_TYPE 0x00
unsigned short ReadADC(unsigned char adc_input)
{
	ADMUX = adc_input | (ADC_VREF_TYPE & 0xff);
	_delay_us(10); // Delay needed for the stabilization of the ADC input voltage
	ADCSRA |= 0x40; // Start the AD conversion
	while ((ADCSRA & 0x10) == 0); // Wait for the AD conversion to complete
	ADCSRA |= 0x10;
	return ADCW;
}

SIGNAL(ANA_COMP_vect) // Analog comparator interrupt for over-current detection
{
	if (ACSR & (1<<ACO)) // AIN0 higher than AIN1, i.e M- is above reference = desat error
		PORTB &= ~DESAT_SIGNAL; // Turn output off (= desat error, failsafe)
	else
		PORTB |= DESAT_SIGNAL; // Turn output on (= desat system happy)
}

SIGNAL(TIM0_OVF_vect) // TIM0 used to generate PWM for sending battery voltage via digital optocoupler
{
	PORTB |= PWM_PIN;
}

SIGNAL(TIM0_COMPA_vect)
{
	PORTB &= ~PWM_PIN;
}

short main(void)
{
	// Setup
	wdt_enable(WDTO_15MS);
	DDRB = DESAT_SIGNAL + PWM_PIN;
	TCCR0A = (0<<COM0A1)
			| (0<<COM0A0) // ORC0A disconnected - we switch pin manually instead!! because we're using a different output pin
			| (0<<COM0B1)
			| (0<<COM0B0) // OCR0B disconnected
			| (0<<WGM01)
			| (0<<WGM00); // 000 = normal port operation (i.e no inherent PWM)
	TCCR0B = (0<<FOC0A)
			| (0<<FOC0B)
			| (0<<WGM02)
			| (0<<CS02)
			| (1<<CS01)
			| (0<<CS00); // 100 for CS02:0 is /256 prescaler.. 001 is no prescaler, 010 is /8 prescaler

	TIMSK0 |= (1<<TOIE0) + (1<<OCIE0A); // Overflow interrupt plus output compare interrupt A

	ADCSRA=0x84; // ADC initialization
	
	ACSR |= (1<<ACIE);// + (1<<ACIS1);// + (1<<ACIS0); // Interrupt on rising edge.. falling edge would be just ACIS1 bit, output toggle is neither bit

	sei();

	while (1)
	{
		wdt_reset(); // Heartbeat for watchdog timer
		
		OCR0A = ReadADC(1)>>2; // Scales 0-1023 ADC down to 0-255 PWM duty

		_delay_ms(10); // Cycle main loop at about 100Hz, no hurry
	}
}


