// ZEVA C-series Motor Controller: Code for main microcontroller
// Open Source version, released under MIT License (see Readme file)
// Last modified by Ian Hooper (ZEVA), October 2021

// Code for ATmega16M1 (also suitable for ATmega32M1, ATmega64m1)
// Fuses: 8Mhz+ external crystal with 0ms startup time, CKDIV8 off, brownout 2.7V

// NOTE: It's the same code for MC600C and MC1000C, but be sure to set the correct MAX_CURRENT below

#define MAX_VOLTAGE 180
#define MAX_CURRENT 600 // 600 or 1000, depending on which controller you're building
#define CAN_BAUD_RATE	250 // Code knows how to do 125, 250 and 500 kbps
#define USE_29BIT_IDS	1	// 1 for yes, 0 for no

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>

#define UNDERVOLTAGE		8	// Minimum acceptable battery B+/- voltage
#define MIN_SUPPLY_VOLTAGE	149 // 8V  (10K:1K divider from 12V supply -> 18.62 ADCs per volt)
#define EEPROM_OFFSET		4	// Arbitrary >0

// Thermistor data: 12Kohm at 20degC, 9K at 27degC, 2.5Kohm at 60degC, 1.2Kohm at 80degC, 90degC probably a bit under 1K
#define	THERMAL_CUTBACK		850 // Resistor divider is 10K lowside, NTC thermistor high side
#define THERMAL_SHUTDOWN	950 // About 100deg, this actually emerges via algorithm, but threshold is used for status LEDs

enum { false, true };

enum { NO_MC, MC600C, MC1000C };

enum { STATUS_PACKET_ID = 50, SET_THROTTLE_ID, RECEIVE_SETTINGS_ID, SEND_SETTINGS_ID };

#define CORE_BROADCAST_STATUS	10 // CAN ID for EVMS Core (V2) status, don't need to know any of its other IDs
enum { CORE_IDLE, CORE_PRECHARGING, CORE_RUNNING, CORE_CHARGING, CORE_SETUP };
unsigned char evmsStatus;

#define EVMS3_BROADCAST_STATUS	30 // CAN ID for EVMS V3 status message

// Enumerated settings
enum { MIN_BATT_VOLTAGE, MAX_MOTOR_VOLTAGE, MAX_MOTOR_CURRENT, MAX_BATT_CURRENT,
	THROTTLE_ALGORITHM, THROTTLE_TYPE, IDLE_VOLTAGE, IDLE_CURRENT, NUM_SETTINGS };
unsigned char settings[NUM_SETTINGS];

unsigned char defaults[NUM_SETTINGS] = {
	UNDERVOLTAGE,		// Min batt voltage
	180,				// Max motor voltage
	MAX_CURRENT/10,		// Max motor current, /10 so we increment in 10s, will never need finer than this
	MAX_CURRENT/10,		// Max batt current, same scaling as above
	1,					// Throttle algorithm - ramp rate 1, linear speed control, linear torque control
	1,					// Throttle type  (1 = analog + enable, 2 = dual analog, 3 = HEPA)
	0,					// Idle voltage (0 for idle off)
	0 };				// Idle current (0 for idle off)

short minimums[NUM_SETTINGS] = { UNDERVOLTAGE, 1, 1, 1, 0, 1, 0, 0 };
short maximums[NUM_SETTINGS] = { 150, 180, MAX_CURRENT/10, MAX_CURRENT/10, 255, 3, 12, 20 };

enum { NO_THROTTLE, THROTTLE_ANALOG_ENABLE, THROTTLE_5KOHM_ENABLE, THROTTLE_HEPA };
enum { EEPROM_CORRECT, EEPROM_BLANK, EEPROM_CORRUPT };
enum { NO_ERROR, SLEEPING, DESAT_ERROR, CURRENT_SENSOR_FAULT, TEMP_SENSOR_FAULT, UNDER_VOLTAGE,
	OVER_VOLTAGE, LOW_LOGIC_VOLTAGE, THROTTLE_ERROR, THERMAL_CUTBACK_ERROR, THERMAL_SHUTDOWN_ERROR };
unsigned char error = NO_ERROR;
unsigned char latchingError = false;

#define	GREEN	(1<<PD2) // Pins used for status LED
#define	RED		(1<<PD3)
#define BLUE	(1<<PD4)

// ADC channels
#define THROT_A		2
#define THROT_B		3
#define AUX_VOLTS	8
#define	TEMP		4
#define ISENSE		7
#define BAT_VOLTS	6

// Function declarations
int GetThrottle();
void WriteDefaultsToEEPROM();
void WriteToEEPROM();
char ReadFromEEPROM();
void SetupPorts();
void SetLED(int col);
inline int Cap(int value, int lower, int upper);
inline long CapLong(long value, long lower, long upper);
void CanTX(long packetID, char bytes);

// Global variable declarations
volatile unsigned int lowPWM = 0;
volatile unsigned char desatError = false;
volatile uint8_t data[8];
volatile uint8_t txData[8];
volatile uint8_t settingsRequested = false;
volatile uint8_t settingsUpdated = false;
volatile unsigned short ticks = 0;

short rampRates[] = { 32, 16, 8, 4 }; // Rate of change for throttle
unsigned char rampRate, speedControl, torqueControl;
enum { LINEAR, SEMIQUADRATIC, QUADRATIC, OFF };

short supplyVoltageSamples[16];
char supplyVoltageDidComeUp = false;
char supplyVoltageError = false;

char enableIdle = false;

short canThrottle;
short canThrottleTimer;
#define CAN_THROTTLE_TIMEOUT 3900 // Quarter of a second

// ADC vals for temps:  0   20   40   60   80  100  120 dec C
short tempData[7] = { 292, 465, 674, 819, 914, 948, 985 };

SIGNAL(TIMER1_OVF_vect) // Interrupt at overflow of main PWM timer
{
	OCR1B = lowPWM; // Set Output Compare Register to the desired PWM duty
	ticks++; // Used for main loop timing
	canThrottleTimer++; // Used for timing out the CAN throttle if haven't received an update for a while
}

SIGNAL(SIG_INTERRUPT1) // Desat signal
{
	DDRC &= ~(1<<PC1); // Turn PWM off instantly (don't wait for main loop)
	desatError = true;
}

SIGNAL(CAN_INT_vect) // Interrupt to process received CAN message
{
	int8_t length, savedCANPage;
	savedCANPage = CANPAGE; // Saves current MOB
	CANPAGE = CANHPMOB & 0xF0; // Selects MOB with highest priority interrupt
	if (CANSTMOB & (1<<RXOK))
	{
		length = CANCDMOB & 0x0F; // Number of bytes to receive is bottom four bits of this reg
		for (int8_t i=0; i<length; i++)
			data[i] = CANMSG; // This autoincrements when read
		
		// ID has top 8 bits in IDT1 and bottom 3 bits at TOP of IDT2
		long rxPacketID;
		if (USE_29BIT_IDS)
			rxPacketID = (((long)CANIDT1)<<21) + (((long)CANIDT2)<<13) + (CANIDT3<<5) + (CANIDT4>>3);
		else
			rxPacketID = (CANIDT1<<3) + (CANIDT2>>5);		
	
		if (rxPacketID == RECEIVE_SETTINGS_ID)
		{
			if (data[0] > 0) // then we're receiving new data
			{
				char allValid = true;
				for (int n=0; n<8; n++)
					if (data[n] < minimums[n] || data[n] > maximums[n])
						allValid = false;

				if (allValid) // Then copy data to settings
				{
					for (int n=0; n<8; n++) settings[n] = data[n];
					settingsUpdated = true; // Set this flag so the main loop will save new settings to EEPROM
				}
			}
			settingsRequested = true; // Set this flag so main loop will transmit current settings
		}
		else if (rxPacketID == SET_THROTTLE_ID)
		{
			canThrottle = data[0];
			canThrottleTimer = 0;
		}
		else if (rxPacketID == CORE_BROADCAST_STATUS)
		{
			evmsStatus = data[0]&0b00000111; // Bottom three bits have status
		}
		else if (rxPacketID == EVMS3_BROADCAST_STATUS)
		{
			evmsStatus = data[0]&0b00000111;
		}

		CANCDMOB = (1<<CONMOB1) | (8<<DLC0) | ((1<<IDE)*USE_29BIT_IDS); // Enable reception, 11-bit IDE, data length 8
		// Note: The DLC field of CANCDMOB register is updated by the received MOB, and if it differs from above, an error is set
	}
	CANSTMOB = 0x00; // Reset interrupt reason on selected channel
	CANPAGE = savedCANPage;
}

#define ADC_VREF_TYPE (1<<REFS0) // AVCC ref with cap on AREF pin
int ReadADC(unsigned char channel)
{
	ADMUX = channel | ADC_VREF_TYPE;
	_delay_us(10);				// Delay needed for the stabilization of the ADC input voltage
	ADCSRA |= (1<<ADSC);				// Start the AD conversion
	while ((ADCSRA & 0x10) == 0);	// Wait for the AD conversion to complete
	ADCSRA |= 0x10;
	return ADCW;
}

inline int Abs(int num)
{
	if (num<0) return -num; else return num;
}

int main(void)
{
	_delay_ms(100); // Allow things to settle
	SetupPorts();
	while (ReadADC(AUX_VOLTS) < MIN_SUPPLY_VOLTAGE) // Low supply voltage lockout
	{
		SetLED(0);
		_delay_ms(200);
		SetLED(BLUE);
		_delay_ms(200);
	}

	for (int n=0; n<16; n++) supplyVoltageSamples[n] = 0; // Initialise averaging buffer

	char eeprom = ReadFromEEPROM();
	if (eeprom == EEPROM_BLANK || eeprom == EEPROM_CORRUPT) // Lost settings warning
	{
		WriteDefaultsToEEPROM(); // Write defaults to EEPROM
		while (1) // Then latched error - important that the user is made aware that settings were corrupt
		{
			SetLED(BLUE);	// Kind of a double-blink
			_delay_ms(100);
			SetLED(0);
			_delay_ms(100);
			SetLED(BLUE);
			_delay_ms(100);
			SetLED(0);
			_delay_ms(500);
		}
	}

	while (GetThrottle() > 0) // High pedal lockout
	{
		SetLED(RED);
		_delay_ms(300);
		SetLED(BLUE);
		_delay_ms(300);
		SetLED(GREEN);
		_delay_ms(300);
	}

	EICRA |= (1<<ISC11);	// INT1 trigger on falling edge - trigger for desat / overcurrent detection
	EIMSK |= (1<<INT1);		// Enable INT1
	sei();

	wdt_enable(WDTO_15MS);

	// Maximum current works out around 60A per TO247 device, which is a bit more than they like,
	// but it's time-shared between the diode and transistor. Current limit is reduced slightly
	// for very small and very large PWM duties, because in these cases either the diodes or the
	// transistors (respectively) are spending almost all the time conducting, and we want to
	// limit their average current to no more than 50 amps per TO-247.
	int envelopeValue = MAX_CURRENT*4/5; // 480 for the 600A, 800 for the 1000A
	int envelopeStep = MAX_CURRENT/250; // 2 for the 600A, 4 for the 1000A

	unsigned int counter = 0;
	int currentPWMx8 = 0; // Some oversampling to make proportional interpolation smoother
	int targetThrottlex8 = 0;// Oversample this one too
	int idleCurrentx8 = 0; // Oversampled so we can do a gentle ramp up of idled motor
	int error = 0;
	int throttleError = false;
	canThrottleTimer = CAN_THROTTLE_TIMEOUT;
	evmsStatus = CORE_RUNNING; // Unless we hear otherwise via CAN, assume we're running
	
	while (1) // Begin infinite loop
	{   
	    wdt_reset();
		counter++;

		int throttle = GetThrottle();

		if (canThrottleTimer < CAN_THROTTLE_TIMEOUT) // i.e received a CAN throttle less than a tenth of a second ago
			throttle = canThrottle*4; // *4 since the CAN throttle comes in as 0-255 range, but we use 0-1023
		else
			canThrottleTimer = CAN_THROTTLE_TIMEOUT; // so it doesn't overflow

		if (throttle == -1) throttleError = true;

		int motorCurrent = (511 - ReadADC(ISENSE))*25/8; // HTFS800 is 800A for 1.25V -> 640A/V -> 3.125A per ADC -> ADC * 25/8
		int temperature = ReadADC(TEMP);
		
		unsigned int batteryVoltage = (1023-ReadADC(BAT_VOLTS))/5 + 2; // 6.8:270 voltage divider, 4.88mV/div -> 200mV/div.. /5 -> 1V resolution
			// Inverted because we're coming through an inverting optocoupler from the ATtiny13A slave
			// Plus 2 because it seems to be a bit out, maybe that won't matter (unless it's a proportionate factor out? but doesn't seem to be?)

		batteryVoltage += batteryVoltage/20; // Seemed to be about 5% out for some reason
		
		if (batteryVoltage < UNDERVOLTAGE) batteryVoltage = 0; // Not valid below UNDERVOLTAGE

		// Use average of supply voltage to make it more noise immune
		supplyVoltageSamples[counter&0xF] = ReadADC(AUX_VOLTS);
		int supplyVoltage = 0;
		for (int n=0; n<16; n++) 
			supplyVoltage += supplyVoltageSamples[n];
		supplyVoltage = supplyVoltage>>4;

		if (supplyVoltage > MIN_SUPPLY_VOLTAGE) supplyVoltageDidComeUp = true;
		if (supplyVoltage < MIN_SUPPLY_VOLTAGE && supplyVoltageDidComeUp) supplyVoltageError = true;
		
        // Status LED (and error latching)
		if (evmsStatus != CORE_RUNNING)
		{
			((counter&0x7FF) < 0x100) ? SetLED(BLUE) : SetLED(0); // Blink blue
			error = SLEEPING;
		}
		else if (batteryVoltage > MAX_VOLTAGE || motorCurrent < -500 || motorCurrent > MAX_CURRENT+200 || temperature < 50 || temperature > 1000 || desatError || latchingError) // temporary
		{
			latchingError = true; // Latch on, it's a fatal error
			if (batteryVoltage > MAX_VOLTAGE) error = OVER_VOLTAGE;
			if (motorCurrent < -500 || motorCurrent > MAX_CURRENT+200) error = CURRENT_SENSOR_FAULT;
			if (temperature<50 || temperature > 1000) error = TEMP_SENSOR_FAULT;
			if (desatError) error = DESAT_ERROR;
			
			if (desatError)
				(counter & 0x100) ? SetLED(RED) : SetLED(0);	// Red/off for desat error
			else
				(counter & 0x200) ? SetLED(RED) : SetLED(BLUE);	// Red/blue for other fatal error
		}
		else if (supplyVoltage < MIN_SUPPLY_VOLTAGE || supplyVoltageError)
		{
			error = LOW_LOGIC_VOLTAGE;
			(counter & 0x100) ? SetLED(BLUE) : SetLED(0); // Flash blue for low supply voltage
		}
		else if (batteryVoltage < settings[MIN_BATT_VOLTAGE]+2)  // +2 so it flickers less at limit
		{
			error = UNDER_VOLTAGE;
			SetLED(BLUE);							// Blue for undervoltage
		}
		else if (throttleError)
		{
			error = THROTTLE_ERROR;
			(counter & 0x100) ? SetLED(GREEN) : SetLED(0); // Flash green for throttle error
		}
		else if (temperature > THERMAL_SHUTDOWN)
		{
			error = THERMAL_SHUTDOWN_ERROR;
			SetLED(RED);							// Red for overtemp
		}
		else if (temperature > THERMAL_CUTBACK+50) // +50 because we don't want to worry people until significantly overtemp
		{
			error = THERMAL_CUTBACK_ERROR;
			(counter & 0x100) ? SetLED(GREEN) : SetLED(RED); // Green/red for thermal cutback
		}
		else
		{
			error = NO_ERROR;
			SetLED(GREEN);	// Green light for all happy
		}

		// Extract throttle algorithm info
		rampRate = Cap(settings[THROTTLE_ALGORITHM] & 0b00001111, 0, 4);
		speedControl = Cap((settings[THROTTLE_ALGORITHM]&0b00110000)>>4, LINEAR, OFF);
		torqueControl = Cap((settings[THROTTLE_ALGORITHM]&0b11000000)>>6, LINEAR, OFF);

		// If we're at zero throttle (and hence no pending throttle error), and a throttleError was flagged previously, reset
		if (throttleError && throttle == 0) throttleError = false;
		if (throttleError || throttle < 0) throttle = 0; // Don't allow any throttle if a throttle error is on (until reset by zero throttle, above)
        
		// Do throttle ramp
		if (rampRate == 0)// No ramp
			targetThrottlex8 = throttle*8;
		else
		{
			if (throttle*8 > targetThrottlex8)
				targetThrottlex8 += rampRates[rampRate-1]; // Ramp up by set speed
			if (throttle*8 < targetThrottlex8)
				targetThrottlex8 = throttle*8; // Instant down, also corrects if ramp up step was bigger than required
		}
		
		// New throttle to voltage (PWM) algorithm
		unsigned int range = 255;
		if (batteryVoltage > settings[MAX_MOTOR_VOLTAGE])
			range = 255 * (unsigned int)settings[MAX_MOTOR_VOLTAGE] / (unsigned int)batteryVoltage;

		// targetThrottlex8 is 0-8184 range, >>6 makes it 0-127
		unsigned int targetThrottleQuadratic = (targetThrottlex8>>6)*(targetThrottlex8>>6)/127;
		
		unsigned int targetSpeedLinear = Cap((targetThrottlex8>>6) * range / 128, 0, 255);
		unsigned int targetSpeedQuadratic = Cap(targetThrottleQuadratic * range / 128, 0, 255);
        
		int tempy = targetThrottlex8>>5; // 0-255
		if (tempy > 0) enableIdle = true; // Need to blip throttle to start idle functionality, so it doesn't go while precharging

		long targetCurrentLinear = (long)settings[MAX_MOTOR_CURRENT] * 10L * (long)tempy / 255L;
		long targetCurrentQuadratic = (long)settings[MAX_MOTOR_CURRENT] * 10L * (long)tempy * (long)tempy / 65025L; // 255 x 255 = 65025

		int targetPWM = 0;
		int targetCurrent = 0;
		if (speedControl != OFF || torqueControl != OFF) // Both off would be invalid, leave target outputs zero
		{
			if (speedControl == LINEAR)
				targetPWM = targetSpeedLinear;
			else if (speedControl == SEMIQUADRATIC)
				targetPWM = (targetSpeedLinear+targetSpeedQuadratic)/2L;
			else if (speedControl == QUADRATIC)
				targetPWM = targetSpeedQuadratic;
			else if (speedControl == OFF && tempy > 0)
				targetPWM = range;

			if (torqueControl == LINEAR)
				targetCurrent = targetCurrentLinear;
			else if (torqueControl == SEMIQUADRATIC)
				targetCurrent = (targetCurrentLinear+targetCurrentQuadratic)/2L;
			else if (torqueControl == QUADRATIC)
				targetCurrent = targetCurrentQuadratic;
			else if (torqueControl == OFF)
				targetCurrent = settings[MAX_MOTOR_CURRENT]*10L;
		}

		if (enableIdle) // Next 5 lines are for motor idle functionality
		{
			int idlePWM = 0;
			if (batteryVoltage > 0 && settings[IDLE_VOLTAGE] > 0) idlePWM = 255 * settings[IDLE_VOLTAGE] / batteryVoltage; // Calc the PWM based on voltage
			if (targetPWM < idlePWM) targetPWM = idlePWM;
			if (idleCurrentx8 < settings[IDLE_CURRENT]*10*8) idleCurrentx8++; // Slowly ramp up idle current limit
			if (targetCurrent < (idleCurrentx8>>3)) targetCurrent = idleCurrentx8>>3;
		}

		// Envelope to limit average current for very high and very low PWMs
		int currentPWM = currentPWMx8>>3;
		targetCurrent = Cap(targetCurrent, 0, envelopeValue + currentPWM * envelopeStep); // 0-50 PWM have 80%-100% current ramp
		targetCurrent = Cap(targetCurrent, 0, envelopeValue + (255-currentPWM) * envelopeStep); // 205-255 PWM have 100%-80% current ramp				
		
		// Thermal limiting - reduce targetCurrent by a percentage based on how far over thermal limit threshold we are
		// We do a two stage ramp for smoother response, 850-900 represents 65-75degC and drops to 75%, 900-950 is 75-100degC and drops to zero
		if (temperature > THERMAL_CUTBACK+50) // 900 to 950
			targetCurrent = targetCurrent * (long)(150-(temperature-(THERMAL_CUTBACK+50))*3)/200L; // Lose 1.5% per ADC for 75% to 0% over 50 ADCs
		else if (temperature > THERMAL_CUTBACK) // 850 to 900
			targetCurrent = targetCurrent * (long)(200-(temperature-THERMAL_CUTBACK))/200L; // Lose 0.5% per ADC to taper 100% to 75%
		
		// Estimation algorithm for target PWM based on current current and PWM
		long targetPWMforMotorCurrent = targetPWM; // [0-255] normally
		if (currentPWMx8 > 8 && motorCurrent > 5)
			targetPWMforMotorCurrent = CapLong((long)(currentPWMx8>>3) * (long)targetCurrent / (long)motorCurrent, 0, 255L);
        
        long batteryCurrent = (long)motorCurrent * (long)(currentPWMx8>>3) / 255L;
        
		long targetBattCurrent = (long)settings[MAX_BATT_CURRENT] * 10L * (long)tempy / 255L;
		long targetPWMforBattCurrent = targetPWM;
		if (currentPWMx8 > 8 && batteryCurrent > 2)
			targetPWMforBattCurrent = CapLong((long)(currentPWMx8>>3) * (long)targetBattCurrent / (long)batteryCurrent, 0, 255L);

		if (targetPWMforBattCurrent < targetPWM) targetPWM = targetPWMforBattCurrent;
		if (targetPWMforMotorCurrent < targetPWM) targetPWM = targetPWMforMotorCurrent; // Choose lowest target current (envelope)

		// Limit minimum battery voltage
        if (batteryVoltage < settings[MIN_BATT_VOLTAGE]) targetPWM = 0; // Crude but functional			
        
		targetPWM = Cap(targetPWM, 0, 255); // Verify correct range

		// Interpolate currentPWM to targetPWM
		int delta = targetPWM - (currentPWMx8>>3);
		currentPWMx8 += delta;

		if (error) currentPWMx8 = 0; // Shouldn't be required but better safe than sorry

		lowPWM = currentPWMx8>>2; // Atomic assignment to volatile variable, 0-511 range

		if (targetPWM == 0 || error || supplyVoltageError || evmsStatus != CORE_RUNNING)
			DDRC &= ~(1<<PC1); // Turn PWM off
		else
			DDRC |= (1<<PC1); // Enable PWM

		// CAN stuff
		if (ticks > 1560) // 10Hz // Every quarter of a second
		{
			ticks -= 1560;

			// Calculate temperature in degC using linear interpolation from array
			int temperatureC = 0; // i.e "not available" if the algorithm below doesn't find it (0-160degC)
			for (int n=0; n<6; n++)
			{
				if (temperature >= tempData[n] && temperature < tempData[n+1]) // we're between samples  
				{
					temperatureC = n*20 + 20*(temperature - tempData[n])/(tempData[n+1] - tempData[n]);
					break;
				}
			}
			temperatureC += 3; // Manual correction for low end of range which is out

			int controllerType = MC600C;
			if (MAX_CURRENT == 1000) controllerType = MC1000C;

			// Status data
			txData[0] = controllerType + (error<<4);
			txData[1] = batteryVoltage;
			txData[2] = Cap(batteryCurrent/5, 0, 255); // So 1000A -> 200, won't overflow U8
			txData[3] = batteryVoltage * (lowPWM/2) / 255; // converts to motorVoltage
			txData[4] = Cap(motorCurrent/5, 0, 255);
			txData[5] = temperatureC;
			txData[6] = (throttle>>2)*100/255; // Internal throttle is 0-1023, so >>2 makes it 0-255, then convert to 0-100% range
			if (batteryVoltage > 255) txData[6] += 0b10000000; // Add 9th bit if it's high voltage
			txData[7] = lowPWM>>1; // Maybe useful to have? 0-255
			CanTX(STATUS_PACKET_ID, 8);
		}

		if (settingsRequested)
		{
			settingsRequested = false;
			for (int n=0; n<8; n++) txData[n] = settings[n];
			CanTX(SEND_SETTINGS_ID, 8);
		}

		if (settingsUpdated)
		{
			settingsUpdated = false;
			wdt_disable();
			WriteToEEPROM(); // Save updated settings
			wdt_enable(WDTO_15MS);
		}
	}
}

int GetThrottle()
{
	int input1 = ReadADC(THROT_A);
	int input2 = ReadADC(THROT_B);

	int throttle = 0;

	if (settings[THROTTLE_TYPE] == THROTTLE_ANALOG_ENABLE)
	{
		if (input1 > 256 && input2 < 512) // More than 25% analog throttle without Enable coming on
			return -1; // return error
		
		if (input2 > 512) // 2nd input pulled high
			throttle = input1;
	}
	else if (settings[THROTTLE_TYPE] == THROTTLE_5KOHM_ENABLE)
	{
		if (input1 < 512 // Would equate to 10Kohm resistance
			|| (input1 < 850 && input2 < 512)) // Significant throttle level without Enable coming on
			return -1; // error
		if (input2 > 512) // Enable high
			throttle = (960-input1)*3; // 960 because schottky diode drops 5V supply down to about 4.7V	
	}
	else if (settings[THROTTLE_TYPE] == THROTTLE_HEPA)
	{
		// HEPA: throttle1 should be 0.7-3.5V, throttle2 1.4-4.2, 0.7v offset
		if (input1 < 100 /* ~0.5V */ || input2 < 200 /* ~1V */ || input1 > 950 /* ~4.6V */ || input2 > 1000 /* ~4.9V */
				|| Abs(input1+128 /* 0.6V */ - input2) > 64 /* 0.3V */) // Throttles too low, too high, or outside diff range of 0.3V to 0.9V
			return -1; //error
		else
			throttle = ((input1-150) + (input2-300))*4/5; // Nominally we want this scaled to 0-1024 at this point, add deadbands later
	}
	else
		return -1; // Unknown throttle type or something? Should never get here, but just in case

	return Cap(throttle*8/7-64, 0, 1023); // Add deadbands and return
}

void SetupPorts()
{
	DDRD = RED+BLUE+GREEN;
	
	// Timer1 does PWM at 16khz, i.e 9-bit = 512 counts at 8Mhz (WGM12 + WGM11)
	TCCR1A |= (1<<COM1B1) + (1<<WGM11); // Clear OC1B on compare match
	TCCR1B |= (1<<WGM12) + (1<<CS10); // Clock running, no prescaler
	TIMSK1 |= (1<<TOIE1); // Interrupt on overflow so we can set next PWM duty cycle
	
	// Set up analog to digital
	ADMUX |= ADC_VREF_TYPE;
	ADCSRA |= 0b10000110; // ADEN plus 110 prescaler = /64 for 125kHz ADC clock
	ADCSRB |= (1<<AREFEN); // WARNING: This seems to screw up the ADC for some reason (slows down settling time or something?)
	
	// CAN bus init stuff
	CANGCON = (1<<SWRES); // Software reset
	CANTCON = 0; // CAN timing prescaler set to 0
	if (CAN_BAUD_RATE == 500)
		CANBT1 = 0x02;
	else if (CAN_BAUD_RATE == 250)
		CANBT1 = 0x06;
	else
		CANBT1 = 0x0E;
	CANBT2 = 0x04;
	CANBT3 = 0x13; // These set baud rate to 125kb/s (see ATmega16M1 manual, page 203)

	for (int8_t mob=0; mob<6; mob++)
	{
		CANPAGE = (mob<<4); // Select MOB 0-5
		CANCDMOB = 0x00; // Disable MOB
		CANSTMOB = 0x00; // Clear MOB status register
	}

	CANPAGE = (1<<MOBNB0); // Select MOB1
	CANIE2 = (1<<IEMOB1); // Enable interrupts on MOB1 for reception and transmission
	CANGIE = (1<<ENIT) | (1<<ENRX); // Enable interrupts on receive and transmit
	CANIDM1 = CANIDM2 = CANIDM3 = CANIDM4 = 0x00; // CAN ID mask, zero will let all IDs pass
	CANCDMOB = (1<<CONMOB1) + (8<<DLC0) + ((1<<IDE)*USE_29BIT_IDS); // Enable transmission, 29-bit ID, 8-bit data
	//CANCDMOB = (1<<CONMOB1) + (8<<DLC0) + ((1<<IDE)*USE_29BIT_IDS);
	CANGCON |= (1<<ENASTB); // Enable mode. CAN channel enters enable mode after 11 recessive bits have been read
}

void SetLED(int col)
{
	PORTD &= ~(RED+GREEN+BLUE); // PD2-4 off
	PORTD |= col;
}

inline int Cap(int value, int lower, int upper)
{
	if (value < lower)
		return lower;
	else if (value > upper)
		return upper;
	else
		return value;
}

inline long CapLong(long value, long lower, long upper)
{
	if (value < lower)
		return lower;
	else if (value > upper)
		return upper;
	else
		return value;
}

void WriteDefaultsToEEPROM()
{
	for (int n=0; n<NUM_SETTINGS; n++)
		settings[n] = defaults[n];
	WriteToEEPROM();
}

void WriteToEEPROM()
{
	cli();	// Disable interrupts
	unsigned char checksum = 0;
	for (int i=0; i<NUM_SETTINGS; i++)
	{
		eeprom_write_byte((unsigned char*)(EEPROM_OFFSET+i), settings[i]);
		checksum += settings[i];
	}
	eeprom_write_byte((unsigned char*)(EEPROM_OFFSET+NUM_SETTINGS), checksum);		// Add checksum
	eeprom_write_dword((unsigned long*)(EEPROM_OFFSET+NUM_SETTINGS+1), 0xC0FFEE);	// Add watermark
	sei();
	
	EEAR = 0; // Park EEPROM pointer at sacrifical location 0
}

char ReadFromEEPROM()
{
	if (eeprom_read_dword((unsigned long*)(EEPROM_OFFSET+NUM_SETTINGS+1)) != 0xC0FFEE)
		return EEPROM_BLANK;

	unsigned char checksum = 0;
	for (int i=0; i<NUM_SETTINGS; i++)
	{
		settings[i] = eeprom_read_byte((unsigned char*)(EEPROM_OFFSET+i));
		checksum += settings[i];
		if (settings[i] < minimums[i] || settings[i] > maximums[i]) return EEPROM_CORRUPT; // If value out of range - probably due to memory corruption
	}

	if (checksum != eeprom_read_byte((unsigned char*)(EEPROM_OFFSET+NUM_SETTINGS))) return EEPROM_CORRUPT;
    
    EEAR = 0; // Park EEPROM pointer at sacrificial location 0 when we're done - this is what usually gets trampled during brownouts etc

	return EEPROM_CORRECT;
}

void CanTX(long packetID, char bytes)
{
	CANPAGE = 0x00; // Select MOB0 for transmission
	while (CANEN2 & (1<<ENMOB0)); // Wait for MOB0 to be free
	CANSTMOB = 0x00;

	if (USE_29BIT_IDS) // CAN 2.0b is 29-bit IDs, CANIDT4 has bits 0-4 in top 5 bits, CANID3 has 5-12
	{
		CANIDT1 = packetID>>21;
		CANIDT2 = packetID>>13;
		CANIDT3 = packetID>>5;
		CANIDT4 = (packetID & 0b00011111)<<3;
	}
	else // CAN 2.0a is 11-bit IDs, IDT1 has top 8 bits, IDT2 has bottom three bits BUT at top of byte!
	{
		CANIDT1 = (packetID>>3); // Packet ID
		CANIDT2 = (packetID & 0x07)<<5;
		CANIDT3 = 0x00;
		CANIDT4 = 0x00;
	}

	for (int8_t i=0; i<bytes; i++) CANMSG = txData[i];

	CANCDMOB = (1<<CONMOB0) | (bytes<<DLC0) | ((1<<IDE)*USE_29BIT_IDS); // Enable transmission, 29-bit ID, 8-bit data

	while (!(CANSTMOB & (1<<TXOK)) && !(CANSTMOB & (1<<AERR))); // Wait for transmission to finish (via setting of TXOK flag)

	CANCDMOB = 0x00; // Disable transmission
	CANSTMOB = 0x00; // Clear TXOK flag
}

